@@ -4,7 +4,7 @@ use warnings;
 use strict;
 use base qw(DynaLoader);
 
-our $VERSION = '0.94';
+our $VERSION = '0.97';
 eval { bootstrap Sys::CpuAffinity $VERSION };
 
 sub import {
@@ -23,11 +23,17 @@ sub _maskToArray {
 }
 
 sub _arrayToMask {
-    my ($mask) = (0,0);
+    my $mask = 0;
     $mask |= (1 << $_) for @_;
     return $mask;
 }
 
+
+
+# does 1<<32 equal 1 or 4294967296 on this system?
+# when it's 1, we need to say 2**$np instead of 1<<$np
+our $_INT32 = !(1 - (1 << 32));
+
 #
 # Development guide:
 #
@@ -80,22 +86,18 @@ sub setAffinity {
   my ($pid, $mask, %flags) = @_; # %flags reserved for future use
   if (ref $mask eq 'ARRAY') {
     $mask = 0;
-    $mask |= (1 << $_) for @{$_[1]};
+    $mask += (2 ** $_) for @{$_[1]};
   }
+  my $np = getNumCpus();
   if ($mask == -1) {
-    my $np = getNumCpus();
     if ($np > 0) {
-      $mask = 0;
-      for my $bit (0..$np-1) {
-	$mask |= 1 << $bit;
-      }
+      $mask = (2 ** $np) - 1;
     }
   }
   if ($mask <= 0) {
     carp "Sys::CpuAffinity: invalid mask $mask in call to setAffinty\n";
     return;
   }
-  my $np = getNumCpus();
 
   # http://www.cpantesters.org/cpan/report/07107190-b19f-3f77-b713-d32bba55d77f
   # 1 << 32 == 1  caused test failure in v0.90
@@ -114,19 +116,18 @@ sub setAffinity {
     }
   }
 
-  return 0
-    || _setAffinity_with_Win32API($pid,$mask)
+  return _setAffinity_with_Win32API($pid,$mask)
+    || _setAffinity_with_xs_win32($pid,$mask)
     || _setAffinity_with_Win32Process($pid,$mask)
     || _setAffinity_with_taskset($pid,$mask)
     || _setAffinity_with_xs_sched_setaffinity($pid,$mask)
     || _setAffinity_with_BSD_Process_Affinity($pid,$mask)
-    || _setAffinity_with_xs_cpuset_setaffinity($pid,$mask) # XXX needs work
+    || _setAffinity_with_xs_cpuset_setaffinity($pid,$mask)  # XXX needs wor
     || _setAffinity_with_xs_processor_bind($pid,$mask)
     || _setAffinity_with_bindprocessor($pid,$mask)
     || _setAffinity_with_cpuset($pid,$mask)
     || _setAffinity_with_pbind($pid,$mask)
-#   || _setAffinity_with_psrset($pid,$mask)       # XXX needs proper test env
-    || _setAffinity_with_xs_win32($pid,$mask)
+#   || _setAffinity_with_psrset($pid,$mask) # XXX needs proper test e
     || 0;
 }
 
@@ -149,7 +150,9 @@ sub getNumCpus() {
     || _getNumCpus_from_psrinfo()
     || _getNumCpus_from_hinv()
     || _getNumCpus_from_hwprefs()
+    || _getNumCpus_from_system_profiler()
     || _getNumCpus_from_prtconf()
+    || _getNumCpus_from_Test_Smoke_SysInfo()
     || _getNumCpus_from_ENV()
     || -1;
 }
@@ -171,12 +174,14 @@ sub _getNumCpus_from_ENV {
 }
 
 sub _getNumCpus_from_Win32API {
-  # GetActiveProcessorCount api function is only supported since Windows 7
+  # GetActiveProcessorCount api function is only supported since Windows 7?
+  # !!! Unfortunately, it also seems to make Windows 7 crash !!!
   return 0 if $^O ne "MSWin32" && $^O ne "cygwin";
   return 0 if !_configModule("Win32::API");
 
   # ALL_PROCESSOR_GROUPS: 0xffff
-  return _win32api("GetActiveProcessorCount", 0xffff) || 0;
+  ### return _win32api("GetActiveProcessorCount", 0xffff) ||
+    0;
 }
 
 our %WIN32_SYSTEM_INFO = ();
@@ -194,6 +199,8 @@ sub _getNumCpus_from_Win32API_System_Info {
         $lpsysinfo_type_avail ? 'LPSYSTEM_INFO' : 'PCHAR';
       $WIN32API{"GetSystemInfo"} = Win32::API->new('kernel32', $proto);
     }
+
+    # does this part break on 64-bit machines?
     my $buffer = chr(0) x 36;
     $WIN32API{"GetSystemInfo"}->Call($buffer);
     ($WIN32_SYSTEM_INFO{"PageSize"},
@@ -361,9 +368,13 @@ sub _getNumCpus_from_hinv {   # NOT TESTED irix
   return 0 if $^O =~ /irix/i;
   return 0 if !_configExternalProgram("hinv");
   my $cmd = _configExternalProgram("hinv");
-  my @result = qx($cmd | grep -i processor 2> /dev/null);
-  my ($result) = grep { /^\d+ .* [Pp]rocessors?/ } @result;
-  return (split /\s+/, $result)[0] || 0;
+
+  # found this in Test::Smoke::SysInfo v0.042 in Test-Smoke-1.43 module
+  my @processor = qx($cmd -c processor);
+  _debug("\"hinv -c processor\" output: ", @processor);
+  my ($cpu_cnt) = grep /\d+.+processors?$/i, @processor;
+  my $ncpu = (split " ", $cpu_cnt)[0];
+  return $ncpu;
 }
 
 
@@ -373,9 +384,29 @@ sub _getNumCpus_from_hwprefs {   # NOT TESTED darwin
   my $cmd = _configExternalProgram("hwprefs");
   my $result = qx($cmd cpu_count 2>/dev/null);
   $result =~ s/\s+$//;
+  _debug("\"$cmd cpu_count\" output: ", $result);
   return $result || 0;
 }
 
+sub _getNumCpus_from_system_profiler {  # NOT TESTED darwin
+  return 0 if $^O !~ /darwin/ && $^O !~ /MacOS/i;
+  return 0 if !_configExternalProgram("system_profiler");
+
+  # with help from Test::Smoke::SysInfo
+  my $cmd = _configExternalProgram("system_profiler");
+  my $system_profiler_output 
+    = qx($cmd -detailLevel mini SPHardwardDataType);
+  my %system_profiler;
+  $system_profiler{uc $1} = $2
+    while $system_profiler_output =~ m/^\s*([\w ]+):\s+(.+)$/gm;
+
+  my $ncpus = $system_profiler{'NUMBER OF CPUS'};
+  if (!defined $ncpus) {
+    $ncpus = $system_profiler{'TOTAL NUMBER OF CORES'};
+  }
+  return $ncpus;
+}
+
 sub _getNumCpus_from_prtconf {    # NOT TESTED
   # solaris has a prtconf command, but I don't think it outputs #cpus.
   return 0 if $^O !~ /aix/i;
@@ -389,6 +420,15 @@ sub _getNumCpus_from_prtconf {    # NOT TESTED
   return $ncpus || 0;
 }
 
+sub _getNumCpus_from_Test_Smoke_SysInfo {   # NOT TESTED
+  return 0 if !_configModule('Test::Smoke::SysInfo');
+  my $sysinfo = Test::Smoke::SysInfo->new();
+  if (defined $sysinfo && defined $sysinfo->{_ncpu}) {
+    return $sysinfo->{_ncpu};
+  }
+  return;
+}
+
 ######################################################################
 
 # get affinity toolbox
@@ -399,8 +439,8 @@ sub _getAffinity_with_Win32API {
   return 0 if !_configModule("Win32::API");
 
   my $pid = $opid;
-  if ($^O eq "cygwin" && defined &Cygwin::pid_to_winpid) {
-    $pid = Cygwin::pid_to_winpid($opid);
+  if ($^O eq "cygwin") {
+    $pid = __pid_to_winpid($opid);
   }
 
   if ($pid > 0) {
@@ -470,8 +510,8 @@ sub _getAffinity_with_Win32Process {
   return 0 if !_configModule("Win32::Process");
   return 0 if $pid < 0;  # pseudo-process / thread id
 
-  if ($^O eq "cygwin" && defined &Cygwin::pid_to_winpid) {
-    $pid = Cygwin::pid_to_winpid($pid);
+  if ($^O eq "cygwin") {
+    $pid = __pid_to_winpid($pid);
   }
 
   my ($processHandle, $processMask, $systemMask, $result);
@@ -523,7 +563,7 @@ sub _getAffinity_with_pbind {
     } else {
       carp "_getAffinity_with_pbind: ",
 	"process $pid unbound but can't count processors\n";
-      return 0xFFFF; 
+      return 2**32 - 1;
     }
   } elsif ($pbind_output =~ /: (\d+)/) {
     my $bound_processor = $1;
@@ -597,8 +637,8 @@ sub _getAffinity_with_xs_cpuset_getaffinity {
 sub _getAffinity_with_xs_win32 {
   my ($opid) = @_;
   my $pid = $opid;
-  if ($^O =~ /cygwin/ && defined &Cygwin::pid_to_winpid) {
-    $pid = Cygwin::pid_to_winpid($opid);
+  if ($^O =~ /cygwin/) {
+    $pid = __pid_to_winpid($opid);
   }
   if ($pid < 0) {
     return 0 if !defined &xs_win32_getAffinity_thread;
@@ -627,8 +667,8 @@ sub _setAffinity_with_Win32API {
 
   # if $^O is "cygwin", make sure you are passing the Windows pid,
   # using Cygwin::pid_to_winpid if necessary!
-  if ($^O eq "cygwin" && defined &Cygwin::pid_to_winpid) {
-    $pid = Cygwin::pid_to_winpid($pid);
+  if ($^O eq "cygwin") {
+    $pid = __pid_to_winpid($pid);
   }
 
   if ($pid > 0) {
@@ -648,7 +688,7 @@ sub _setAffinity_with_Win32API {
     # 0x0200: THREAD_SET_LIMITED_INFORMATION
     my $threadHandle;
     local $! = undef;
-    $^E = '';
+    $^E = 0;
     return 0 unless $threadHandle
       = _win32api("OpenThread", 0x0060, 0, -$pid)
 	|| _win32api("OpenThread", 0x0600, 0, -$pid)
@@ -668,16 +708,23 @@ sub _setAffinity_with_Win32Process {
   my ($pid, $mask) = @_;
   return 0 if $^O ne "MSWin32" && $^O ne "cygwin";
   return 0 if !_configModule("Win32::Process");
-  if ($^O eq "cygwin" && defined &Cygwin::pid_to_winpid) {
-    $pid = Cygwin::pid_to_winpid($pid);
+  if ($^O eq "cygwin") {
+    $pid = __pid_to_winpid($pid);
   }
 
   my $processHandle;
   return 0 unless Win32::Process::Open($processHandle, $pid, 0)
     && ref $processHandle eq 'Win32::Process';
-  my $result = $processHandle->SetProcessAffinityMask($mask);
-  _debug("set affinity with Win32::Process: $result");
-  return $result;
+
+  eval {
+    # Seg fault on Cygwin? We really prefer not to use it on Cygwin.
+    local $SIG{SEGV} = 'IGNORE';
+    return 0 if $^O eq 'cygwin';
+    my $result = $processHandle->SetProcessAffinityMask($mask);
+    _debug("set affinity with Win32::Process: $result");
+    return $result;
+  };
+  return 0;
 }
 
 sub _setAffinity_with_taskset {
@@ -740,7 +787,12 @@ sub _setAffinity_with_xs_processor_bind {
   } else {
     my @amask = _maskToArray($mask);
     return 0 if !defined &xs_setaffinity_process_bind;
-    my $result = xs_setaffinity_processor_bind($pid,$amask[0]);
+
+    # solaris processor_bind() is for binding to a single processor.
+    # see comment under _setAffinity_with_pbind().
+
+    my $element = 0;
+    my $result = xs_setaffinity_processor_bind($pid,$amask[$element]);
     _debug("result from setaffinity_processor_bind: $result");
     return $result;
   }
@@ -806,7 +858,7 @@ sub _setAffinity_with_xs_win32 {
   my ($opid, $mask) = @_;
   my $pid = $opid;
   if ($^O =~ /cygwin/) {
-    $pid = Cygwin::pid_to_winpid($opid);
+    $pid = __pid_to_winpid($opid);
   }
 
   if ($pid < 0) {
@@ -832,6 +884,30 @@ sub _setAffinity_with_xs_win32 {
   return 0;
 }
 
+sub __pid_to_winpid {
+  my ($cygwinpid) = @_;
+  if ($] >= 5.008 && defined(&Cygwin::pid_to_winpid)) {
+    return Cygwin::pid_to_winpid($cygwinpid);
+  } else {
+    return __poor_mans_pid_to_winpid($cygwinpid);
+  }
+}
+
+sub __poor_mans_pid_to_winpid {
+  my ($cygwinpid) = @_;
+  my @psw = `/usr/bin/ps -W`;
+  foreach my $psw (@psw) {
+    $psw =~ s/^[A-Z\s]+//;
+    my ($pid,$ppid,$pgid,$winpid) = split /\s+/, $psw;
+    next unless $pid;
+    if ($pid == $cygwinpid) {
+      return $winpid;
+    }
+  }
+  warn "Could not resolve cygwin pid $cygwinpid into winpid.\n";
+  return $cygwinpid;
+}
+
 ######################################################################
 
 # configuration code
@@ -871,6 +947,13 @@ sub _configExternalProgram {
   if ($^O ne 'MSWin32') {
     my $which = qx(which $program 2>/dev/null);
     $which =~ s/\s+$//;
+
+    if ($which =~ / not in / 			# negative output on irix
+	|| $which =~ /no \Q$program\E in /	# negative output on solaris
+       ) {
+
+      $which = '';
+    }
     if ($which) {
       _debug("Program $program is available in $which");
       return $PROGRAM{$program} = $which;
@@ -948,7 +1031,7 @@ sub _win32api {
       return;
     }
   }
-  return if !defined $WIN32API{$function} || $WIN32API{$function} == 0;
+  return if !defined($WIN32API{$function}) || $WIN32API{$function} == 0;
   return $WIN32API{$function}->Call(@_);
 }
 
@@ -966,7 +1049,7 @@ Sys::CpuAffinity - Set CPU affinity for processes
 
 =head1 VERSION
 
-Version 0.94
+Version 0.97
 
 =head1 SYNOPSIS
 
@@ -1015,21 +1098,23 @@ several techniques for manipulating CPU affinities in
 other existing modules, and Sys::CpuAffinity will use
 these modules if they are available:
 
-    Inline, Inline::C [useful for any system]
     Win32::API, Win32::Process [MSWin32, cygwin]
     BSD::Process::Affinity [*bsd]
-    Solaris::Lgrp [solaris]
 
 =head1 SUPPORTED SYSTEMS
 
 The techniques for manipulating CPU affinities for Windows
 (including Cygwin) and Linux have been refined and tested
 pretty well. Some techniques applicable to BSD systems
-(particularly FreeBSD) have been tested a little bit. The
-hope is that this module will include more techniques for
+(particularly FreeBSD) and Solaris have been tested a little bit. 
+The hope is that this module will include more techniques for
 more systems in future releases. See the L</"NOTE TO DEVELOPERS">
 below for information about how you can help.
 
+MacOS is explicitly not supported, as there does not appear to
+be any public interface for specifying the CPU affinity of
+a process directly.
+
 =head1 SUBROUTINES/METHODS
 
 =over 4
@@ -1051,7 +1136,7 @@ So for example, if a process in an 8-CPU machine
 had affinity for CPU's # 2, 6, and 7, then
 in scalar context, C<getAffinity()> would return
 
-    (1 << 2) || (1 << 6) | (1 << 7) ==> 196
+    (1 << 2) | (1 << 6) | (1 << 7) ==> 196
 
 and in array context, it would return
 
@@ -1153,9 +1238,11 @@ be included in the next release.
 
 =head1 ACKNOWLEDGEMENTS
 
-=cut 
+L<BSD::Process::Affinity> for demonstrating how to get/set affinities
+on BSD systems.
 
-# Some BSD code shamelessly lifted from L<BSD::Process::Affinity>.
+L<Test::Smoke::SysInfo> has some fairly portable code for detecting
+the number of processors.
 
 =head1 LICENSE AND COPYRIGHT
 
@@ -1223,6 +1310,10 @@ Rumors of cpu affinity on other systems:
     AIX:  /bindprocessor, bindprocessor() in <sys/processor.h>
     MacOS: thread_policy_set(),thread_policy_get() in <mach/thread_policy.h>
 
+	In MacOS it is possible to assign threads to the same
+	processor, but generally not to assign them to any particular
+	processor. MacOS is totally unsupported for now.
+
 
 how to find the number of processors:
     AIX:  sysconf(_SC_NPROCESSORS_CONF), sysconf(_SC_NPROCESSORS_ONLN)
@@ -1249,3 +1340,8 @@ by default.
 Some systems have a concept of the maximum number of processors that
 they can suppport.
 
+Currently (0.91-0.97), constant parameters to Win32 API functions are 
+hard coded, not extracted from the local header files. Microsoft is
+probably loathe to change these constants between different versions,
+but this still seems dodgy.
+
